@@ -1,35 +1,36 @@
 /* modpriv.h: Stuff needed by both modules.c and modcall.c, but should not be
  * accessed from anywhere else.
  *
- * Version: $Id$ */
+ * Version: $Id: d5e2c2392bac5f7aa5ea74eccba53bd9861ab328 $ */
 #ifndef FR_MODPRIV_H
 #define FR_MODPRIV_H
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
 
-#ifndef WITHOUT_LIBLTDL
-#include "ltdl.h"
+#ifndef HAVE_DLFCN_H
+#error FreeRADIUS needs either libltdl, or a working dlopen()
 #else
+#include <dlfcn.h>
+#endif
+
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 typedef void *lt_dlhandle;
 
-lt_dlhandle lt_dlopenext(const char *name);
-void *lt_dlsym(lt_dlhandle handle, const char *symbol);
+lt_dlhandle lt_dlopenext(char const *name);
+void *lt_dlsym(lt_dlhandle handle, char const *symbol);
 int lt_dlclose(lt_dlhandle handle);
-const char *lt_dlerror(void);
-
-#define LTDL_SET_PRELOADED_SYMBOLS(_x)
-#define lt_dlinit(_x) (0)
-#define lt_dlexit(_x)
-#define lt_dlsetsearchpath(_x)
-#endif
+char const *lt_dlerror(void);
 
 /*
  *	Keep track of which modules we've loaded.
  */
 typedef struct module_entry_t {
 	char			name[MAX_STRING_LEN];
-	const module_t		*module;
+	module_t const		*module;
 	lt_dlhandle		handle;
 } module_entry_t;
 
@@ -43,16 +44,26 @@ typedef struct fr_module_hup_t fr_module_hup_t;
 typedef struct module_instance_t {
 	char			name[MAX_STRING_LEN];
 	module_entry_t		*entry;
-	void                    *insthandle;
+	void			*insthandle;
 #ifdef HAVE_PTHREAD_H
 	pthread_mutex_t		*mutex;
 #endif
 	CONF_SECTION		*cs;
-	int			dead;
+	time_t			last_hup;
+	bool			instantiated;
+	bool			force;
+	rlm_rcode_t		code;
 	fr_module_hup_t	       	*mh;
 } module_instance_t;
 
-module_instance_t *find_module_instance(CONF_SECTION *, const char *instname,
-					int do_link);
-int module_hup_module(CONF_SECTION *cs, module_instance_t *node, time_t when);
+module_instance_t	*module_instantiate(CONF_SECTION *modules, char const *askedname);
+module_instance_t	*module_instantiate_method(CONF_SECTION *modules, char const *askedname, rlm_components_t *method);
+module_instance_t	*module_find(CONF_SECTION *modules, char const *askedname);
+int			find_module_sibling_section(CONF_SECTION **out, CONF_SECTION *module, char const *name);
+int			module_hup_module(CONF_SECTION *cs, module_instance_t *node, time_t when);
+
+#ifdef __cplusplus
+}
+#endif
+
 #endif	/* FR_MODPRIV_H */
