@@ -2,7 +2,7 @@
  * event.c	Non-thread-safe event handling, specific to a RADIUS
  *		server.
  *
- * Version:	$Id$
+ * Version:	$Id: fa1e0aa0dafdcfac6c188111288cc6be3522d6ff $
  *
  *   This library is free software; you can redistribute it and/or
  *   modify it under the terms of the GNU Lesser General Public
@@ -22,13 +22,21 @@
  *  Copyright 2007  Alan DeKok <aland@ox.org>
  */
 
-#include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: fa1e0aa0dafdcfac6c188111288cc6be3522d6ff $")
 
 #include <freeradius-devel/libradius.h>
 #include <freeradius-devel/heap.h>
 #include <freeradius-devel/event.h>
 
+#ifdef HAVE_KQUEUE
+#ifndef HAVE_SYS_EVENT_H
+#error kqueue requires <sys/event.h>
+
+#else
+#include <sys/event.h>
+#endif
+#endif	/* HAVE_KQUEUE */
+
 typedef struct fr_event_fd_t {
 	int			fd;
 	fr_event_fd_handler_t	handler;
@@ -36,22 +44,30 @@ typedef struct fr_event_fd_t {
 } fr_event_fd_t;
 
 #define FR_EV_MAX_FDS (256)
+
 #undef USEC
 #define USEC (1000000)
 
 struct fr_event_list_t {
 	fr_heap_t	*times;
 
-	int		changed;
-
 	int		exit;
 
 	fr_event_status_t status;
 
 	struct timeval  now;
-	int		dispatch;
+	bool		dispatch;
 
+	int		num_readers;
+#ifndef HAVE_KQUEUE
 	int		max_readers;
+
+	bool		changed;
+
+#else
+	int		kq;
+	struct kevent	events[FR_EV_MAX_FDS]; /* so it doesn't go on the stack every time */
+#endif
 	fr_event_fd_t	readers[FR_EV_MAX_FDS];
 };
 
@@ -62,15 +78,15 @@ struct fr_event_t {
 	fr_event_callback_t	callback;
 	void			*ctx;
 	struct timeval		when;
-	fr_event_t		**ev_p;
+	fr_event_t		**parent;
 	int			heap;
 };
 
 
-static int fr_event_list_time_cmp(const void *one, const void *two)
+static int fr_event_list_time_cmp(void const *one, void const *two)
 {
-	const fr_event_t *a = one;
-	const fr_event_t *b = two;
+	fr_event_t const *a = one;
+	fr_event_t const *b = two;
 
 	if (a->when.tv_sec < b->when.tv_sec) return -1;
 	if (a->when.tv_sec > b->when.tv_sec) return +1;
@@ -82,34 +98,39 @@ static int fr_event_list_time_cmp(const void *one, const void *two)
 }
 
 
-void fr_event_list_free(fr_event_list_t *el)
+static int _event_list_free(fr_event_list_t *list)
 {
+	fr_event_list_t *el = list;
 	fr_event_t *ev;
 
-	if (!el) return;
-
 	while ((ev = fr_heap_peek(el->times)) != NULL) {
 		fr_event_delete(el, &ev);
 	}
 
 	fr_heap_delete(el->times);
-	free(el);
+
+#ifdef HAVE_KQUEUE
+	close(el->kq);
+#endif
+
+	return 0;
 }
 
 
-fr_event_list_t *fr_event_list_create(fr_event_status_t status)
+fr_event_list_t *fr_event_list_create(TALLOC_CTX *ctx, fr_event_status_t status)
 {
 	int i;
 	fr_event_list_t *el;
 
-	el = malloc(sizeof(*el));
-	if (!el) return NULL;
-	memset(el, 0, sizeof(*el));
+	el = talloc_zero(ctx, fr_event_list_t);
+	if (!fr_assert(el)) {
+		return NULL;
+	}
+	talloc_set_destructor(el, _event_list_free);
 
-	el->times = fr_heap_create(fr_event_list_time_cmp, 
-				   offsetof(fr_event_t, heap));
+	el->times = fr_heap_create(fr_event_list_time_cmp, offsetof(fr_event_t, heap));
 	if (!el->times) {
-		fr_event_list_free(el);
+		talloc_free(el);
 		return NULL;
 	}
 
@@ -117,12 +138,29 @@ fr_event_list_t *fr_event_list_create(fr_event_status_t status)
 		el->readers[i].fd = -1;
 	}
 
+#ifndef HAVE_KQUEUE
+	el->changed = true;	/* force re-set of fds's */
+
+#else
+	el->kq = kqueue();
+	if (el->kq < 0) {
+		talloc_free(el);
+		return NULL;
+	}
+#endif
+
 	el->status = status;
-	el->changed = 1;	/* force re-set of fds's */
 
 	return el;
 }
 
+int fr_event_list_num_fds(fr_event_list_t *el)
+{
+	if (!el) return 0;
+
+	return el->num_readers;
+}
+
 int fr_event_list_num_elements(fr_event_list_t *el)
 {
 	if (!el) return 0;
@@ -131,49 +169,96 @@ int fr_event_list_num_elements(fr_event_list_t *el)
 }
 
 
-int fr_event_delete(fr_event_list_t *el, fr_event_t **ev_p)
+int fr_event_delete(fr_event_list_t *el, fr_event_t **parent)
 {
+	int ret;
+
 	fr_event_t *ev;
 
-	if (!el || !ev_p || !*ev_p) return 0;
+	if (!el || !parent || !*parent) return 0;
 
-	ev = *ev_p;
-	if (ev->ev_p) *(ev->ev_p) = NULL;
-	*ev_p = NULL;
+#ifndef NDEBUG
+	/*
+	 *  Validate the event_t struct to detect memory issues early.
+	 */
+	ev = talloc_get_type_abort(*parent, fr_event_t);
 
-	fr_heap_extract(el->times, ev);
-	free(ev);
+#else
+	ev = *parent;
+#endif
 
-	return 1;
+	if (ev->parent) {
+		fr_assert(*(ev->parent) == ev);
+		*ev->parent = NULL;
+	}
+	*parent = NULL;
+
+	ret = fr_heap_extract(el->times, ev);
+	fr_assert(ret == 1);	/* events MUST be in the heap */
+	talloc_free(ev);
+
+	return ret;
 }
 
 
-int fr_event_insert(fr_event_list_t *el,
-		      fr_event_callback_t callback,
-		      void *ctx, struct timeval *when,
-		      fr_event_t **ev_p)
+int fr_event_insert(fr_event_list_t *el, fr_event_callback_t callback, void *ctx, struct timeval *when,
+		    fr_event_t **parent)
 {
 	fr_event_t *ev;
 
-	if (!el || !callback | !when || (when->tv_usec > USEC)) return 0;
+	if (!el) {
+		fr_strerror_printf("Invalid arguments (NULL event list)");
+		return 0;
+	}
 
-	if (ev_p && *ev_p) fr_event_delete(el, ev_p);
+	if (!callback) {
+		fr_strerror_printf("Invalid arguments (NULL callback)");
+		return 0;
+	}
 
-	ev = malloc(sizeof(*ev));
-	if (!ev) return 0;
-	memset(ev, 0, sizeof(*ev));
+	if (!when || (when->tv_usec >= USEC)) {
+		fr_strerror_printf("Invalid arguments (time)");
+		return 0;
+	}
+
+	if (!parent) {
+		fr_strerror_printf("Invalid arguments (NULL parent)");
+		return 0;
+	}
+
+	/*
+	 *	If there is an event, re-use it instead of freeing it
+	 *	and allocating a new one.
+	 */
+	if (*parent) {
+		int ret;
+
+#ifndef NDEBUG
+		ev = talloc_get_type_abort(*parent, fr_event_t);
+#else
+		ev = *parent;
+#endif
+
+		ret = fr_heap_extract(el->times, ev);
+		fr_assert(ret == 1);	/* events MUST be in the heap */
+
+		memset(ev, 0, sizeof(*ev));
+	} else {
+		ev = talloc_zero(el, fr_event_t);
+		if (!ev) return 0;
+	}
 
 	ev->callback = callback;
 	ev->ctx = ctx;
 	ev->when = *when;
-	ev->ev_p = ev_p;
+	ev->parent = parent;
 
 	if (!fr_heap_insert(el->times, ev)) {
-		free(ev);
+		talloc_free(ev);
 		return 0;
 	}
 
-	if (ev_p) *ev_p = ev;
+	*parent = ev;
 	return 1;
 }
 
@@ -199,6 +284,10 @@ int fr_event_run(fr_event_list_t *el, struct timeval *when)
 		return 0;
 	}
 
+#ifndef NDEBUG
+	ev = talloc_get_type_abort(ev, fr_event_t);
+#endif
+
 	/*
 	 *	See if it's time to do this one.
 	 */
@@ -215,7 +304,7 @@ int fr_event_run(fr_event_list_t *el, struct timeval *when)
 	/*
 	 *	Delete the event before calling it.
 	 */
-	fr_event_delete(el, &ev);
+	fr_event_delete(el, ev->parent);
 
 	callback(ctx);
 	return 1;
@@ -242,13 +331,77 @@ int fr_event_fd_insert(fr_event_list_t *el, int type, int fd,
 	int i;
 	fr_event_fd_t *ef;
 
-	if (!el || (fd < 0) || !handler || !ctx) return 0;
+	if (!el) {
+		fr_strerror_printf("Invalid arguments (NULL event list)");
+		return 0;
+	}
 
-	if (type != 0) return 0;
+	if (!handler) {
+		fr_strerror_printf("Invalid arguments (NULL handler)");
+		return 0;
+	}
 
-	if (el->max_readers >= FR_EV_MAX_FDS) return 0;
+	if (!ctx) {
+		fr_strerror_printf("Invalid arguments (NULL ctx)");
+		return 0;
+	}
 
+	if (fd < 0) {
+		fr_strerror_printf("Invalid arguments (bad FD %i)", fd);
+		return 0;
+	}
+
+	if (type != 0) {
+		fr_strerror_printf("Invalid type %i", type);
+		return 0;
+	}
+
+	if (el->num_readers >= FR_EV_MAX_FDS) {
+		fr_strerror_printf("Too many readers");
+		return 0;
+	}
 	ef = NULL;
+
+#ifdef HAVE_KQUEUE
+	/*
+	 *	We need to store TWO fields with the event.  kqueue
+	 *	only lets us store one.  If we put the two fields into
+	 *	a malloc'd structure, that would help.  Except that
+	 *	kqueue can silently delete the event when the socket
+	 *	is closed, and not give us the opportunity to free it.
+	 *	<sigh>
+	 *
+	 *	The solution is to put the fields into an array, and
+	 *	do a linear search on addition/deletion of the FDs.
+	 *	However, to avoid MOST linear issues, we start off the
+	 *	search at "FD" offset.  Since FDs are unique, AND
+	 *	usually less than 256, we do "FD & 0xff", which is a
+	 *	good guess, and makes the lookups mostly O(1).
+	 */
+	for (i = 0; i < FR_EV_MAX_FDS; i++) {
+		int j;
+		struct kevent evset;
+
+		j = (i + fd) & (FR_EV_MAX_FDS - 1);
+
+		if (el->readers[j].fd >= 0) continue;
+
+		/*
+		 *	We want to read from the FD.
+		 */
+		EV_SET(&evset, fd, EVFILT_READ, EV_ADD | EV_ENABLE, 0, 0, &el->readers[j]);
+		if (kevent(el->kq, &evset, 1, NULL, 0, NULL) < 0) {
+			fr_strerror_printf("Failed inserting event for FD %i: %s", fd, fr_syserror(errno));
+			return 0;
+		}
+
+		ef = &el->readers[j];
+		el->num_readers++;
+		break;
+	}
+
+#else  /* HAVE_KQUEUE */
+
 	for (i = 0; i <= el->max_readers; i++) {
 		/*
 		 *	Be fail-safe on multiple inserts.
@@ -256,6 +409,7 @@ int fr_event_fd_insert(fr_event_list_t *el, int type, int fd,
 		if (el->readers[i].fd == fd) {
 			if ((el->readers[i].handler != handler) ||
 			    (el->readers[i].ctx != ctx)) {
+				fr_strerror_printf("Multiple handlers for same FD");
 				return 0;
 			}
 
@@ -267,19 +421,26 @@ int fr_event_fd_insert(fr_event_list_t *el, int type, int fd,
 
 		if (el->readers[i].fd < 0) {
 			ef = &el->readers[i];
+			el->num_readers++;
 
 			if (i == el->max_readers) el->max_readers = i + 1;
 			break;
 		}
 	}
+#endif
 
-	if (!ef) return 0;
+	if (!ef) {
+		fr_strerror_printf("Failed assigning FD");
+		return 0;
+	}
 
+	ef->fd = fd;
 	ef->handler = handler;
 	ef->ctx = ctx;
-	ef->fd = fd;
 
-	el->changed = 1;
+#ifndef HAVE_KQUEUE
+	el->changed = true;
+#endif
 
 	return 1;
 }
@@ -292,17 +453,47 @@ int fr_event_fd_delete(fr_event_list_t *el, int type, int fd)
 
 	if (type != 0) return 0;
 
+#ifdef HAVE_KQUEUE
+	for (i = 0; i < FR_EV_MAX_FDS; i++) {
+		int j;
+		struct kevent evset;
+
+		j = (i + fd) & (FR_EV_MAX_FDS - 1);
+
+		if (el->readers[j].fd != fd) continue;
+
+		/*
+		 *	Tell the kernel to delete it from the list.
+		 *
+		 *	The caller MAY have closed it, in which case
+		 *	the kernel has removed it from the list.  So
+		 *	we ignore the return code from kevent().
+		 */
+		EV_SET(&evset, fd, EVFILT_READ, EV_DELETE, 0, 0, NULL);
+		(void) kevent(el->kq, &evset, 1, NULL, 0, NULL);
+
+		el->readers[j].fd = -1;
+		el->num_readers--;
+
+		return 1;
+	}
+
+#else
+
 	for (i = 0; i < el->max_readers; i++) {
 		if (el->readers[i].fd == fd) {
 			el->readers[i].fd = -1;
+			el->num_readers--;
+
 			if ((i + 1) == el->max_readers) el->max_readers = i;
-			el->changed = 1;
+			el->changed = true;
 			return 1;
 		}
 	}
+#endif	/* HAVE_KQUEUE */
 
 	return 0;
-}			 
+}
 
 
 void fr_event_loop_exit(fr_event_list_t *el, int code)
@@ -312,35 +503,50 @@ void fr_event_loop_exit(fr_event_list_t *el, int code)
 	el->exit = code;
 }
 
+bool fr_event_loop_exiting(fr_event_list_t *el)
+{
+	return (el->exit != 0);
+}
 
 int fr_event_loop(fr_event_list_t *el)
 {
-	int i, rcode, maxfd = 0;
+	int i, rcode;
 	struct timeval when, *wake;
+#ifdef HAVE_KQUEUE
+	struct timespec ts_when, *ts_wake;
+#else
+	int maxfd = 0;
 	fd_set read_fds, master_fds;
 
+	el->changed = true;
+#endif
+
 	el->exit = 0;
-	el->dispatch = 1;
-	el->changed = 1;
+	el->dispatch = true;
 
 	while (!el->exit) {
+#ifndef HAVE_KQUEUE
 		/*
 		 *	Cache the list of FD's to watch.
 		 */
 		if (el->changed) {
+#ifdef __clang_analyzer__
+			memset(&master_fds, 0, sizeof(master_fds));
+#else
 			FD_ZERO(&master_fds);
-			
+#endif
 			for (i = 0; i < el->max_readers; i++) {
 				if (el->readers[i].fd < 0) continue;
-				
+
 				if (el->readers[i].fd > maxfd) {
 					maxfd = el->readers[i].fd;
 				}
 				FD_SET(el->readers[i].fd, &master_fds);
 			}
-			
-			el->changed = 0;
+
+			el->changed = false;
 		}
+#endif	/* HAVE_KQUEUE */
 
 		/*
 		 *	Find the first event.  If there's none, we wait
@@ -353,7 +559,9 @@ int fr_event_loop(fr_event_list_t *el)
 			fr_event_t *ev;
 
 			ev = fr_heap_peek(el->times);
-			if (!ev) _exit(42);
+			if (!ev) {
+				fr_exit_now(42);
+			}
 
 			gettimeofday(&el->now, NULL);
 
@@ -364,9 +572,11 @@ int fr_event_loop(fr_event_list_t *el)
 				if (when.tv_sec > 0) {
 					when.tv_sec--;
 					when.tv_usec += USEC;
+				} else {
+					when.tv_sec = 0;
 				}
 				when.tv_usec -= el->now.tv_usec;
-				if (when.tv_usec > USEC) {
+				if (when.tv_usec >= USEC) {
 					when.tv_usec -= USEC;
 					when.tv_sec++;
 				}
@@ -385,36 +595,87 @@ int fr_event_loop(fr_event_list_t *el)
 		 */
 		if (el->status) el->status(wake);
 
+#ifndef HAVE_KQUEUE
 		read_fds = master_fds;
 		rcode = select(maxfd + 1, &read_fds, NULL, NULL, wake);
 		if ((rcode < 0) && (errno != EINTR)) {
-			el->dispatch = 0;
-			return 0;
+			fr_strerror_printf("Failed in select: %s", fr_syserror(errno));
+			el->dispatch = false;
+			return -1;
 		}
 
+#else  /* HAVE_KQUEUE */
+
+		if (wake) {
+			ts_wake = &ts_when;
+			ts_when.tv_sec = when.tv_sec;
+			ts_when.tv_nsec = when.tv_usec * 1000;
+		} else {
+			ts_wake = NULL;
+		}
+
+		rcode = kevent(el->kq, NULL, 0, el->events, FR_EV_MAX_FDS, ts_wake);
+#endif	/* HAVE_KQUEUE */
+
 		if (fr_heap_num_elements(el->times) > 0) {
 			do {
 				gettimeofday(&el->now, NULL);
 				when = el->now;
 			} while (fr_event_run(el, &when) == 1);
 		}
-		
+
 		if (rcode <= 0) continue;
 
+#ifndef HAVE_KQUEUE
+		/*
+		 *	Loop over all of the sockets to see if there's
+		 *	an event for that socket.
+		 */
 		for (i = 0; i < el->max_readers; i++) {
 			fr_event_fd_t *ef = &el->readers[i];
 
 			if (ef->fd < 0) continue;
 
 			if (!FD_ISSET(ef->fd, &read_fds)) continue;
-			
+
 			ef->handler(el, ef->fd, ef->ctx);
 
 			if (el->changed) break;
 		}
+
+#else  /* HAVE_KQUEUE */
+
+		/*
+		 *	Loop over all of the events, servicing them.
+		 */
+		for (i = 0; i < rcode; i++) {
+			fr_event_fd_t *ef = el->events[i].udata;
+
+			if (el->events[i].flags & EV_EOF) {
+				/*
+				 *	FIXME: delete the handler
+				 *	here, and fix process.c to not
+				 *	call fr_event_fd_delete().
+				 *	It's cleaner.
+				 *
+				 *	Call the handler, which SHOULD
+				 *	delete the connection.
+				 */
+				ef->handler(el, ef->fd, ef->ctx);
+				continue;
+			}
+
+			/*
+			 *	Else it's our event.  We only set
+			 *	EVFILT_READ, so it must be a read
+			 *	event.
+			 */
+			ef->handler(el, ef->fd, ef->ctx);
+		}
+#endif	/* HAVE_KQUEUE */
 	}
 
-	el->dispatch = 0;
+	el->dispatch = false;
 	return el->exit;
 }
 
@@ -468,7 +729,7 @@ int main(int argc, char **argv)
 	struct timeval now, when;
 	fr_event_list_t *el;
 
-	el = fr_event_list_create();
+	el = fr_event_list_create(NULL, NULL);
 	if (!el) exit(1);
 
 	memset(&rand_pool, 0, sizeof(rand_pool));
@@ -503,7 +764,7 @@ int main(int argc, char **argv)
 		}
 	}
 
-	fr_event_list_free(el);
+	talloc_free(el);
 
 	return 0;
 }
