@@ -1,7 +1,7 @@
 /*
  * rlm_eap.h    Local Header file.
  *
- * Version:     $Id$
+ * Version:     $Id: 0edf462bd83bf6cbdc45a334741e4179edfa596c $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -24,8 +24,7 @@
 #ifndef _RLM_EAP_H
 #define _RLM_EAP_H
 
-#include <freeradius-devel/ident.h>
-RCSIDH(rlm_eap_h, "$Id$")
+RCSIDH(rlm_eap_h, "$Id: 0edf462bd83bf6cbdc45a334741e4179edfa596c $")
 
 #include <freeradius-devel/modpriv.h>
 #include "eap.h"
@@ -34,13 +33,13 @@ RCSIDH(rlm_eap_h, "$Id$")
 /*
  * Keep track of which sub modules we've loaded.
  */
-typedef struct eap_types_t {
-	const char	*typename;
-	EAP_TYPE       	*type;
-	lt_dlhandle     handle;
-	CONF_SECTION	*cs;
-	void		*type_data;
-} EAP_TYPES;
+typedef struct eap_module {
+	char const		*name;
+	rlm_eap_module_t	*type;
+	lt_dlhandle		handle;
+	CONF_SECTION		*cs;
+	void			*instance;
+} eap_module_t;
 
 /*
  * This structure contains eap's persistent data.
@@ -48,28 +47,30 @@ typedef struct eap_types_t {
  * types = All supported EAP-Types
  * mutex = ensure only one thread is updating the sessions[] struct
  */
-typedef struct rlm_eap_t {
+typedef struct rlm_eap {
 	rbtree_t	*session_tree;
-	EAP_HANDLER	*session_head, *session_tail;
-	rbtree_t	*handler_tree; /* for debugging only */
-	EAP_TYPES 	*types[PW_EAP_MAX_TYPES + 1];
+	eap_handler_t	*session_head, *session_tail;
+	eap_module_t 	*methods[PW_EAP_MAX_TYPES];
 
 	/*
 	 *	Configuration items.
 	 */
-	int		timer_limit;
-	char		*default_eap_type_name;
-	int		default_eap_type;
-	int		ignore_unknown_eap_types;
-	int		cisco_accounting_username_bug;
-	int		max_sessions;
+	uint32_t	timer_limit;
+
+	char const	*default_method_name;
+	eap_type_t	default_method;
+
+	bool		ignore_unknown_types;
+	bool		mod_accounting_username_bug;
+
+	uint32_t	max_sessions;
 
 #ifdef HAVE_PTHREAD_H
 	pthread_mutex_t	session_mutex;
 	pthread_mutex_t	handler_mutex;
 #endif
 
-	const char	*xlat_name; /* no xlat's yet */
+	char const	*xlat_name; /* no xlat's yet */
 	fr_randctx	rand_pool;
 } rlm_eap_t;
 
@@ -88,28 +89,22 @@ typedef struct rlm_eap_t {
 
 /* function definitions */
 /* EAP-Type */
-int      	eaptype_load(EAP_TYPES **type, int eap_type, CONF_SECTION *cs);
-int       	eaptype_select(rlm_eap_t *inst, EAP_HANDLER *h);
-void		eaptype_free(EAP_TYPES *tl);
+int      	eap_module_instantiate(rlm_eap_t *inst, eap_module_t **method, eap_type_t num, CONF_SECTION *cs);
+eap_rcode_t	eap_method_select(rlm_eap_t *inst, eap_handler_t *handler);
 
 /* EAP */
-int  		eap_start(rlm_eap_t *inst, REQUEST *request);
-void 		eap_fail(EAP_HANDLER *handler);
-void 		eap_success(EAP_HANDLER *handler);
-int 		eap_compose(EAP_HANDLER *handler);
-EAP_HANDLER 	*eap_handler(rlm_eap_t *inst, eap_packet_t **eap_msg, REQUEST *request);
+int  		eap_start(rlm_eap_t *inst, REQUEST *request) CC_HINT(nonnull);
+void 		eap_fail(eap_handler_t *handler) CC_HINT(nonnull);
+void 		eap_success(eap_handler_t *handler) CC_HINT(nonnull);
+rlm_rcode_t 	eap_compose(eap_handler_t *handler) CC_HINT(nonnull);
+eap_handler_t 	*eap_handler(rlm_eap_t *inst, eap_packet_raw_t **eap_msg, REQUEST *request) CC_HINT(nonnull);
 
 /* Memory Management */
-EAP_PACKET  	*eap_packet_alloc(void);
-EAP_DS      	*eap_ds_alloc(void);
-EAP_HANDLER 	*eap_handler_alloc(rlm_eap_t *inst);
-void	    	eap_packet_free(EAP_PACKET **eap_packet);
+EAP_DS      	*eap_ds_alloc(eap_handler_t *handler);
+eap_handler_t 	*eap_handler_alloc(rlm_eap_t *inst);
 void	    	eap_ds_free(EAP_DS **eap_ds);
-void	    	eap_handler_free(rlm_eap_t *inst, EAP_HANDLER *handler);
-
-int 	    	eaplist_add(rlm_eap_t *inst, EAP_HANDLER *handler);
-EAP_HANDLER 	*eaplist_find(rlm_eap_t *inst, REQUEST *request,
-			      eap_packet_t *eap_packet);
+int 	    	eaplist_add(rlm_eap_t *inst, eap_handler_t *handler) CC_HINT(nonnull);
+eap_handler_t 	*eaplist_find(rlm_eap_t *inst, REQUEST *request, eap_packet_raw_t *eap_packet);
 void		eaplist_free(rlm_eap_t *inst);
 
 /* State */
