@@ -1,10 +1,8 @@
 /*
- * rlm_passwd.c
- *
- *   This program is free software; you can redistribute it and/or modify
+ *   This program is is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
- *   the Free Software Foundation; either version 2 of the License, or
- *   (at your option) any later version.
+ *   the Free Software Foundation; either version 2 of the License, or (at
+ *   your option) any later version.
  *
  *   This program is distributed in the hope that it will be useful,
  *   but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -14,15 +12,20 @@
  *   You should have received a copy of the GNU General Public License
  *   along with this program; if not, write to the Free Software
  *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
- *
- * Copyright 2000,2006  The FreeRADIUS server project
  */
 
-#include <freeradius-devel/ident.h>
-RCSID("$Id$")
+/**
+ * $Id: e9957d23d3a56042fbabc6a73042ea4dfdf07047 $
+ * @file rlm_passwd.c
+ * @brief Enables authentication against unix passwd files.
+ *
+ * @copyright 2000,2006  The FreeRADIUS server project
+ */
+RCSID("$Id: e9957d23d3a56042fbabc6a73042ea4dfdf07047 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
+#include <freeradius-devel/rad_assert.h>
 
 struct mypasswd {
 	struct mypasswd *next;
@@ -38,7 +41,6 @@ struct hashtable {
 	int ignorenis;
 	char * filename;
 	struct mypasswd **table;
-	struct mypasswd *last_found;
 	char buffer[1024];
 	FILE *fp;
 	char delimiter;
@@ -50,31 +52,31 @@ struct hashtable {
 #define rad_malloc(s) malloc(s)
 
 void printpw(struct mypasswd *pw, int nfields){
-  int i;
-  if (pw) {
-  	for( i = 0; i < nfields; i++ ) printf("%s:", pw->field[i]);
-  	printf("\n");
-  }
-  else printf ("Not found\n");
-  fflush(stdout);
+	int i;
+	if (pw) {
+		for( i = 0; i < nfields; i++ ) printf("%s:", pw->field[i]);
+		printf("\n");
+	}
+	else printf ("Not found\n");
+	fflush(stdout);
 }
 #endif
 
 
-static struct mypasswd * mypasswd_malloc(const char* buffer, int nfields, size_t* len)
+static struct mypasswd * mypasswd_malloc(char const* buffer, int nfields, size_t* len)
 {
 	struct mypasswd *t;
 	/* reserve memory for (struct mypasswd) + listflag (nfields * sizeof (char*)) +
 	** fields (nfields * sizeof (char)) + strlen (inst->format) + 1 */
 
-  	*len=sizeof (struct mypasswd) + nfields * sizeof (char*) + nfields * sizeof (char ) + strlen(buffer) + 1;
+	*len=sizeof (struct mypasswd) + nfields * sizeof (char*) + nfields * sizeof (char ) + strlen(buffer) + 1;
 	t = (struct mypasswd *) rad_malloc(*len);
 	if (t) memset(t, 0, *len);
-  	return (t);
+	return (t);
 }
 
-static int string_to_entry(const char* string, int nfields, char delimiter,
-	struct mypasswd *passwd, size_t bufferlen)
+static int string_to_entry(char const* string, int nfields, char delimiter,
+			   struct mypasswd *passwd, size_t bufferlen)
 {
 	char *str;
 	size_t len, i;
@@ -119,7 +121,7 @@ static void destroy_password (struct mypasswd * pass)
 }
 
 
-static unsigned int hash(const char * username, unsigned int tablesize)
+static unsigned int hash(char const * username, unsigned int tablesize)
 {
 	int h=1;
 	while (*username) {
@@ -131,10 +133,10 @@ static unsigned int hash(const char * username, unsigned int tablesize)
 static void release_hash_table(struct hashtable * ht){
 	int i;
 
-	if (ht == NULL) return;
+	if (!ht) return;
 	for (i = 0; i < ht->tablesize; i++)
- 		if (ht->table[i])
- 			destroy_password(ht->table[i]);
+		if (ht->table[i])
+			destroy_password(ht->table[i]);
 	if (ht->table) {
 		free(ht->table);
 		ht->table = NULL;
@@ -149,15 +151,16 @@ static void release_hash_table(struct hashtable * ht){
 static void release_ht(struct hashtable * ht){
 	if (!ht) return;
 	release_hash_table(ht);
-	if (ht->filename) free(ht->filename);
+	if (ht->filename) {
+		free(ht->filename);
+		ht->filename = NULL;
+	}
 	free(ht);
 }
 
-static struct hashtable * build_hash_table (const char * file, int nfields,
-	int keyfield, int islist, int tablesize, int ignorenis, char delimiter)
+static struct hashtable * build_hash_table (char const * file, int nfields,
+					    int keyfield, int islist, int tablesize, int ignorenis, char delimiter)
 {
-#define passwd ((struct mypasswd *) ht->buffer)
-	char buffer[1024];
 	struct hashtable* ht;
 	size_t len;
 	unsigned int h;
@@ -165,6 +168,7 @@ static struct hashtable * build_hash_table (const char * file, int nfields,
 	char *list;
 	char *nextlist=0;
 	int i;
+	char buffer[1024];
 
 	ht = (struct hashtable *) rad_malloc(sizeof(struct hashtable));
 	if(!ht) {
@@ -187,8 +191,14 @@ static struct hashtable * build_hash_table (const char * file, int nfields,
 	if(!(ht->fp = fopen(file,"r"))) {
 		free(ht->filename);
 		free(ht);
-		               return NULL;
+		return NULL;
 	}
+
+	/*
+	 *	@todo: This code is SHIT.  It's badly formatted.  It's
+	 *	hard to understand.  It re-implements tons of things
+	 *	which are already in the server core.
+	 */
 	memset(ht->buffer, 0, 1024);
 	ht->table = (struct mypasswd **) rad_malloc (tablesize * sizeof(struct mypasswd *));
 	if (!ht->table) {
@@ -245,21 +255,21 @@ static struct hashtable * build_hash_table (const char * file, int nfields,
 #undef passwd
 }
 
-static struct mypasswd * get_next(char *name, struct hashtable *ht)
+static struct mypasswd * get_next(char *name, struct hashtable *ht,
+				  struct mypasswd **last_found)
 {
-#define passwd ((struct mypasswd *) ht->buffer)
+	struct mypasswd * passwd;
 	struct mypasswd * hashentry;
 	char buffer[1024];
-	int len;
 	char *list, *nextlist;
 
 	if (ht->tablesize > 0) {
 		/* get saved address of next item to check from buffer */
-		hashentry = ht->last_found;
+		hashentry = *last_found;
 		for (; hashentry; hashentry = hashentry->next) {
 			if (!strcmp(hashentry->field[ht->keyfield], name)) {
 				/* save new address */
-				ht->last_found = hashentry->next;
+				*last_found = hashentry->next;
 				return hashentry;
 			}
 		}
@@ -267,9 +277,12 @@ static struct mypasswd * get_next(char *name, struct hashtable *ht)
 	}
 	/*	printf("try to find in file\n"); */
 	if (!ht->fp) return NULL;
+
+	passwd = (struct mypasswd *) ht->buffer;
+
 	while (fgets(buffer, 1024,ht->fp)) {
-		if(*buffer && *buffer!='\n' && (len = string_to_entry(buffer, ht->nfields, ht->delimiter, passwd, sizeof(ht->buffer)-1)) &&
-			(!ht->ignorenis || (*buffer !='-' && *buffer != '+') ) ){
+		if(*buffer && *buffer!='\n' && string_to_entry(buffer, ht->nfields, ht->delimiter, passwd, sizeof(ht->buffer)-1) &&
+		   (!ht->ignorenis || (*buffer !='-' && *buffer != '+') ) ){
 			if(!ht->islist) {
 				if(!strcmp(passwd->field[ht->keyfield], name))
 					return passwd;
@@ -277,9 +290,14 @@ static struct mypasswd * get_next(char *name, struct hashtable *ht)
 			else {
 				for (list = passwd->field[ht->keyfield], nextlist = list; nextlist; list = nextlist) {
 					for(nextlist = list; *nextlist && *nextlist!=','; nextlist++);
-					if(!*nextlist)nextlist = 0;
-					else *nextlist++ = 0;
-					if(!strcmp(list, name)) return passwd;
+					if(!*nextlist) {
+						nextlist = 0;
+					} else {
+						*nextlist++ = 0;
+					}
+					if (!strcmp(list, name)) {
+						return passwd;
+					}
 				}
 			}
 
@@ -288,24 +306,26 @@ static struct mypasswd * get_next(char *name, struct hashtable *ht)
 	fclose(ht->fp);
 	ht->fp = NULL;
 	return NULL;
-#undef passwd
 }
 
-static struct mypasswd * get_pw_nam(char * name, struct hashtable* ht)
+static struct mypasswd * get_pw_nam(char * name, struct hashtable* ht,
+				    struct mypasswd **last_found)
 {
 	int h;
 	struct mypasswd * hashentry;
 
 	if (!ht || !name || *name == '\0') return NULL;
-	ht->last_found = NULL;
+	*last_found = NULL;
 	if (ht->tablesize > 0) {
 		h = hash (name, ht->tablesize);
-		for (hashentry = ht->table[h]; hashentry; hashentry = hashentry->next)
+		for (hashentry = ht->table[h]; hashentry; hashentry = hashentry->next) {
 			if (!strcmp(hashentry->field[ht->keyfield], name)){
 				/* save address of next item to check into buffer */
-				ht->last_found=hashentry->next;
+				*last_found=hashentry->next;
 				return hashentry;
 			}
+		}
+
 		return NULL;
 	}
 	if (ht->fp) {
@@ -313,7 +333,7 @@ static struct mypasswd * get_pw_nam(char * name, struct hashtable* ht)
 		ht->fp = NULL;
 	}
 	if (!(ht->fp=fopen(ht->filename, "r"))) return NULL;
-	return get_next(name, ht);
+	return get_next(name, ht, last_found);
 }
 
 #ifdef TEST
@@ -321,94 +341,91 @@ static struct mypasswd * get_pw_nam(char * name, struct hashtable* ht)
 #define MALLOC_CHECK_ 1
 
 int main(void){
- struct hashtable *ht;
- char *buffer;
- struct mypasswd* pw;
- int i;
+	struct hashtable *ht;
+	char *buffer;
+	struct mypasswd* pw, *last_found;
+	int i;
 
- ht = build_hash_table("/etc/group", 4, 3, 1, 100, 0, ":");
- if(!ht) {
- 	printf("Hash table not built\n");
- 	return -1;
- }
- for (i=0; i<ht->tablesize; i++) if (ht->table[i]) {
-  printf("%d:\n", i);
-  for(pw=ht->table[i]; pw; pw=pw->next) printpw(pw, 4);
- }
+	ht = build_hash_table("/etc/group", 4, 3, 1, 100, 0, ":");
+	if(!ht) {
+		printf("Hash table not built\n");
+		return -1;
+	}
+	for (i = 0; i < ht->tablesize; i++) {
+		if (ht->table[i]) {
+			printf("%d:\n", i);
+			for (pw = ht->table[i]; pw; pw = pw->next) {
+				printpw(pw, 4);
+			}
+		}
+	}
 
- while(fgets(buffer, 1024, stdin)){
-  buffer[strlen(buffer)-1] = 0;
-  pw = get_pw_nam(buffer, ht);
-  printpw(pw,4);
-  while (pw = get_next(buffer, ht)) printpw(pw,4);
- }
- release_ht(ht);
+	while(fgets(buffer, 1024, stdin)){
+		buffer[strlen(buffer)-1] = 0;
+		pw = get_pw_nam(buffer, ht, &last_found);
+		printpw(pw,4);
+		while ((pw = get_next(buffer, ht, &last_found))) printpw(pw,4);
+	}
+	release_ht(ht);
 }
 
 #else  /* TEST */
-struct passwd_instance {
-	struct hashtable *ht;
-	struct mypasswd *pwdfmt;
-	char *filename;
-	char *format;
-	char * delimiter;
-	int allowmultiple;
-	int ignorenislike;
-	int hashsize;
-	int nfields;
-	int keyfield;
-	int listable;
-	int keyattr;
-	int keyattrtype;
-	int ignoreempty;
-};
+typedef struct rlm_passwd_t {
+	struct hashtable	*ht;
+	struct mypasswd		*pwdfmt;
+	char const		*filename;
+	char const		*format;
+	char const		*delimiter;
+	bool			allow_multiple;
+	bool			ignore_nislike;
+	uint32_t		hash_size;
+	uint32_t		nfields;
+	uint32_t		keyfield;
+	uint32_t		listable;
+	DICT_ATTR const		*keyattr;
+	bool			ignore_empty;
+} rlm_passwd_t;
 
 static const CONF_PARSER module_config[] = {
-	{ "filename",   PW_TYPE_FILENAME,
-	   offsetof(struct passwd_instance, filename), NULL,  NULL },
-	{ "format",   PW_TYPE_STRING_PTR,
-	   offsetof(struct passwd_instance, format), NULL,  NULL },
-	{ "delimiter",   PW_TYPE_STRING_PTR,
-	   offsetof(struct passwd_instance, delimiter), NULL,  ":" },
-	{ "ignorenislike",   PW_TYPE_BOOLEAN,
-	   offsetof(struct passwd_instance, ignorenislike), NULL,  "yes" },
-	{ "ignoreempty",   PW_TYPE_BOOLEAN,
-	   offsetof(struct passwd_instance, ignoreempty), NULL,  "yes" },
-	{ "allowmultiplekeys",   PW_TYPE_BOOLEAN,
-	   offsetof(struct passwd_instance, allowmultiple), NULL,  "no" },
-	{ "hashsize",   PW_TYPE_INTEGER,
-	   offsetof(struct passwd_instance, hashsize), NULL,  "100" },
-	{ NULL, -1, 0, NULL, NULL }
+	{ "filename", FR_CONF_OFFSET(PW_TYPE_FILE_INPUT | PW_TYPE_REQUIRED, rlm_passwd_t, filename), NULL },
+	{ "format", FR_CONF_OFFSET(PW_TYPE_STRING | PW_TYPE_REQUIRED, rlm_passwd_t, format), NULL },
+	{ "delimiter", FR_CONF_OFFSET(PW_TYPE_STRING, rlm_passwd_t, delimiter), ":" },
+
+	{ "ignorenislike", FR_CONF_OFFSET(PW_TYPE_BOOLEAN | PW_TYPE_DEPRECATED, rlm_passwd_t, ignore_nislike), NULL },
+	{ "ignore_nislike", FR_CONF_OFFSET(PW_TYPE_BOOLEAN, rlm_passwd_t, ignore_nislike), "yes" },
+
+	{ "ignoreempty", FR_CONF_OFFSET(PW_TYPE_BOOLEAN | PW_TYPE_DEPRECATED, rlm_passwd_t, ignore_empty), NULL },
+	{ "ignore_empty",  FR_CONF_OFFSET(PW_TYPE_BOOLEAN, rlm_passwd_t, ignore_empty), "yes" },
+
+	{ "allowmultiplekeys", FR_CONF_OFFSET(PW_TYPE_BOOLEAN | PW_TYPE_DEPRECATED, rlm_passwd_t, allow_multiple), NULL },
+	{ "allow_multiple_keys", FR_CONF_OFFSET(PW_TYPE_BOOLEAN, rlm_passwd_t, allow_multiple), "no" },
+
+	{ "hashsize", FR_CONF_OFFSET(PW_TYPE_INTEGER | PW_TYPE_DEPRECATED, rlm_passwd_t, hash_size), NULL },
+	{ "hash_size", FR_CONF_OFFSET(PW_TYPE_INTEGER, rlm_passwd_t, hash_size), "100" },
+	CONF_PARSER_TERMINATOR
 };
 
-static int passwd_instantiate(CONF_SECTION *conf, void **instance)
+static int mod_instantiate(CONF_SECTION *conf, void *instance)
 {
-#define inst ((struct passwd_instance *)*instance)
 	int nfields=0, keyfield=-1, listable=0;
-	char *s;
+	char const *s;
 	char *lf=NULL; /* destination list flags temporary */
 	size_t len;
 	int i;
-	DICT_ATTR * da;
+	DICT_ATTR const * da;
+	rlm_passwd_t *inst = instance;
 
-	*instance = rad_malloc(sizeof(struct passwd_instance));
-	if ( !*instance) {
-		radlog(L_ERR, "rlm_passwd: cann't alloc instance");
+	rad_assert(inst->filename && *inst->filename);
+	rad_assert(inst->format && *inst->format);
+
+	if (inst->hash_size == 0) {
+		cf_log_err_cs(conf, "Invalid value '0' for hash_size");
 		return -1;
 	}
-	memset(*instance, 0, sizeof(struct passwd_instance));
-	if (cf_section_parse(conf, inst, module_config) < 0) {
-		free(inst);
-		radlog(L_ERR, "rlm_passwd: cann't parse configuration");
-		return -1;
-	}
-	if(!inst->filename || *inst->filename == '\0' || !inst->format || *inst->format == '\0') {
-		radlog(L_ERR, "rlm_passwd: cann't find passwd file and/or format in configuration");
-		return -1;
-	}
-	lf=strdup(inst->format);
-	if ( lf == NULL) {
-		radlog(L_ERR, "rlm_passwd: memory allocation failed for lf");
+
+	lf = talloc_typed_strdup(inst, inst->format);
+	if ( !lf) {
+		ERROR("rlm_passwd: memory allocation failed for lf");
 		return -1;
 	}
 	memset(lf, 0, strlen(inst->format));
@@ -436,31 +453,30 @@ static int passwd_instantiate(CONF_SECTION *conf, void **instance)
 		s++;
 	}while(*s);
 	if(keyfield < 0) {
-		radlog(L_ERR, "rlm_passwd: no field market as key in format: %s", inst->format);
-		free(lf);
+		cf_log_err_cs(conf, "no field marked as key in format: %s",
+			      inst->format);
 		return -1;
 	}
-	if (! (inst->ht = build_hash_table (inst->filename, nfields, keyfield, listable, inst->hashsize, inst->ignorenislike, *inst->delimiter)) ){
-		radlog(L_ERR, "rlm_passwd: can't build hashtable from passwd file");
-		free(lf);
+	if (! (inst->ht = build_hash_table (inst->filename, nfields, keyfield, listable, inst->hash_size, inst->ignore_nislike, *inst->delimiter)) ){
+		ERROR("rlm_passwd: can't build hashtable from passwd file");
 		return -1;
 	}
 	if (! (inst->pwdfmt = mypasswd_malloc(inst->format, nfields, &len)) ){
-		radlog(L_ERR, "rlm_passwd: memory allocation failed");
+		ERROR("rlm_passwd: memory allocation failed");
 		release_ht(inst->ht);
-		free(lf);
+		inst->ht = NULL;
 		return -1;
 	}
 	if (!string_to_entry(inst->format, nfields, ':', inst->pwdfmt , len)) {
-		radlog(L_ERR, "rlm_passwd: unable to convert format entry");
+		ERROR("rlm_passwd: unable to convert format entry");
 		release_ht(inst->ht);
-		free(lf);
+		inst->ht = NULL;
 		return -1;
 	}
 
 	memcpy(inst->pwdfmt->listflag, lf, nfields);
 
-	free(lf);
+	talloc_free(lf);
 	for (i=0; i<nfields; i++) {
 		if (*inst->pwdfmt->field[i] == '*') inst->pwdfmt->field[i]++;
 		if (*inst->pwdfmt->field[i] == ',') inst->pwdfmt->field[i]++;
@@ -468,106 +484,114 @@ static int passwd_instantiate(CONF_SECTION *conf, void **instance)
 		if (*inst->pwdfmt->field[i] == '~') inst->pwdfmt->field[i]++;
 	}
 	if (!*inst->pwdfmt->field[keyfield]) {
-		radlog(L_ERR, "rlm_passwd: key field is empty");
+		cf_log_err_cs(conf, "key field is empty");
 		release_ht(inst->ht);
+		inst->ht = NULL;
 		return -1;
 	}
 	if (! (da = dict_attrbyname (inst->pwdfmt->field[keyfield])) ) {
-		radlog(L_ERR, "rlm_passwd: unable to resolve attribute: %s", inst->pwdfmt->field[keyfield]);
+		ERROR("rlm_passwd: unable to resolve attribute: %s", inst->pwdfmt->field[keyfield]);
 		release_ht(inst->ht);
+		inst->ht = NULL;
 		return -1;
 	}
-	inst->keyattr = da->attr;
-	inst->keyattrtype = da->type;
+	inst->keyattr = da;
 	inst->nfields = nfields;
 	inst->keyfield = keyfield;
 	inst->listable = listable;
-	DEBUG("rlm_passwd: nfields: %d keyfield %d(%s) listable: %s", nfields, keyfield, inst->pwdfmt->field[keyfield], listable?"yes":"no");
+	DEBUG2("rlm_passwd: nfields: %d keyfield %d(%s) listable: %s", nfields, keyfield, inst->pwdfmt->field[keyfield], listable?"yes":"no");
 	return 0;
 
 #undef inst
 }
 
-static int passwd_detach (void *instance) {
-#define inst ((struct passwd_instance *)instance)
-	if(inst->ht) release_ht(inst->ht);
-	free(instance);
+static int mod_detach (void *instance) {
+#define inst ((rlm_passwd_t *)instance)
+	if(inst->ht) {
+		release_ht(inst->ht);
+		inst->ht = NULL;
+	}
+	free(inst->pwdfmt);
 	return 0;
 #undef inst
 }
 
-static void addresult (struct passwd_instance * inst, REQUEST *request, VALUE_PAIR ** vp, struct mypasswd * pw, char when, const char *listname)
+static void addresult (TALLOC_CTX *ctx, rlm_passwd_t *inst, REQUEST *request,
+		       VALUE_PAIR **vps, struct mypasswd * pw, char when, char const *listname)
 {
-	int i;
-	VALUE_PAIR *newpair;
+	uint32_t i;
+	VALUE_PAIR *vp;
 
-	for (i=0; i<inst->nfields; i++) {
+	for (i = 0; i < inst->nfields; i++) {
 		if (inst->pwdfmt->field[i] && *inst->pwdfmt->field[i] && pw->field[i] && i != inst->keyfield  && inst->pwdfmt->listflag[i] == when) {
-			if ( !inst->ignoreempty || pw->field[i][0] != 0 ) { /* if value in key/value pair is not empty */
-				if (! (newpair = pairmake (inst->pwdfmt->field[i], pw->field[i], T_OP_EQ))) {
-					radlog(L_AUTH, "rlm_passwd: Unable to create %s: %s", inst->pwdfmt->field[i], pw->field[i]);
-					return;
+			if ( !inst->ignore_empty || pw->field[i][0] != 0 ) { /* if value in key/value pair is not empty */
+				vp = fr_pair_make(ctx, vps, inst->pwdfmt->field[i], pw->field[i], T_OP_EQ);
+				if (vp) {
+					RDEBUG("Added %s: '%s' to %s ", inst->pwdfmt->field[i], pw->field[i], listname);
 				}
-				RDEBUG("Added %s: '%s' to %s ", inst->pwdfmt->field[i], pw->field[i], listname);
-				pairadd (vp, newpair);
 			} else
 				RDEBUG("NOOP %s: '%s' to %s ", inst->pwdfmt->field[i], pw->field[i], listname);
 		}
 	}
 }
 
-static int passwd_map(void *instance, REQUEST *request)
+static rlm_rcode_t CC_HINT(nonnull) mod_passwd_map(void *instance, REQUEST *request)
 {
-#define inst ((struct passwd_instance *)instance)
+#define inst ((rlm_passwd_t *)instance)
 	char buffer[1024];
-	VALUE_PAIR * key;
-	struct mypasswd * pw;
-	int found = 0;
+	VALUE_PAIR *key, *i;
+	struct mypasswd * pw, *last_found;
+	vp_cursor_t cursor;
 
-	for (key = request->packet->vps;
-	 key && (key = pairfind (key, inst->keyattr));
-	  key = key->next ){
+	key = fr_pair_find_by_da(request->packet->vps, inst->keyattr, TAG_ANY);
+	if (!key) {
+		return RLM_MODULE_NOTFOUND;
+	}
+
+	for (i = fr_cursor_init(&cursor, &key);
+	     i;
+	     i = fr_cursor_next_by_num(&cursor, inst->keyattr->attr, inst->keyattr->vendor, TAG_ANY)) {
 		/*
 		 *	Ensure we have the string form of the attribute
 		 */
-		vp_prints_value(buffer, sizeof(buffer), key, 0);
-		if (! (pw = get_pw_nam(buffer, inst->ht)) ) {
+		vp_prints_value(buffer, sizeof(buffer), i, 0);
+		if (!(pw = get_pw_nam(buffer, inst->ht, &last_found)) ) {
 			continue;
 		}
 		do {
-			addresult(inst, request, &request->config_items, pw, 0, "config_items");
-			addresult(inst, request, &request->reply->vps, pw, 1, "reply_items");
-			addresult(inst, request, &request->packet->vps, 	pw, 2, "request_items");
-		} while ( (pw = get_next(buffer, inst->ht)) );
-		found++;
-		if (!inst->allowmultiple) break;
-	}
-	if(!found) {
-		return RLM_MODULE_NOTFOUND;
+			addresult(request, inst, request, &request->config, pw, 0, "config");
+			addresult(request->reply, inst, request, &request->reply->vps, pw, 1, "reply_items");
+			addresult(request->packet, inst, request, &request->packet->vps, pw, 2, "request_items");
+		} while ((pw = get_next(buffer, inst->ht, &last_found)));
+
+		if (!inst->allow_multiple) {
+			break;
+		}
 	}
+
 	return RLM_MODULE_OK;
 
 #undef inst
 }
 
+extern module_t rlm_passwd;
 module_t rlm_passwd = {
-	RLM_MODULE_INIT,
-	"passwd",
-	RLM_TYPE_CHECK_CONFIG_SAFE | RLM_TYPE_HUP_SAFE,   	/* type */
-	passwd_instantiate,		/* instantiation */
-	passwd_detach,			/* detach */
-	{
-		NULL,			/* authentication */
-		passwd_map,		/* authorization */
-		NULL,			/* pre-accounting */
-		passwd_map,		/* accounting */
-		NULL,			/* checksimul */
-		NULL,			/* pre-proxy */
-		NULL,			/* post-proxy */
-		passwd_map	       	/* post-auth */
+	.magic		= RLM_MODULE_INIT,
+	.name		= "passwd",
+	.type		= RLM_TYPE_HUP_SAFE,
+	.inst_size	= sizeof(rlm_passwd_t),
+	.config		= module_config,
+	.instantiate	= mod_instantiate,
+	.detach		= mod_detach,
+	.methods = {
+		[MOD_AUTHORIZE]		= mod_passwd_map,
+		[MOD_ACCOUNTING]	= mod_passwd_map,
+		[MOD_POST_AUTH]		= mod_passwd_map,
+		[MOD_PRE_PROXY]		= mod_passwd_map,
+		[MOD_POST_PROXY]  	= mod_passwd_map,
 #ifdef WITH_COA
-		, passwd_map,
-		passwd_map
+		[MOD_RECV_COA]		= mod_passwd_map,
+		[MOD_SEND_COA]		= mod_passwd_map
 #endif
 	},
 };
